#pragma once

#include "common.h"
#include "Globals.h"
#include "ScopedPacketAccess.h"

/** PacketContext contains all relevant dimensional and scaling information for a data packet (specifically one that is derived from WMAreaPacket).
 ** Resolution, worldspace area, and handy functions to convert back and worth between heightfield-space and world-space are all here.
 ** @ingroup data
*/
class PacketContext {
public:
	enum SPACETYPE { WORLD, SPECIFIED, LOCAL };

	PacketContext();
	PacketContext(const PacketContext &other);
	PacketContext(const Rect4F &extents, const Coord &res, const SPACETYPE type = WORLD);

	/// @name Set and Retrieve functions @{
	void set(const Rect4F &region, const int res);
	/// Set the resolution
	void	setRes(const int res);
	/// Set the worldspace render quadrant
	void	setRenderQuad(const Rect4F region);
	/// Set the vertical scale
	void	setVScale(const CoordF vscale);
	
	/// Set X/Y resolution to a specific value with no respect for aspect ratio, etc.
	/// Generally don't use this function unless you know what you're doing!
	/// Since this does not force the region aspect ratio to match the heightfield aspect ratio, you must
	/// use with care unless you are trying to do something funny.
	void	forceRes(const Coord res);

	Coord	getRes() const { return resolution; };
	int		getResMax() const { return (resolution.x > resolution.y) ? resolution.x : resolution.y; };
	inline int		w() const { return resolution.x; };
	inline int		h() const { return resolution.y; };
	CoordF	getRegionSize() const;

	/// Retrieve the origin (which is also the bottom-left corner 
	CoordF		getRenderExtentOrigin() const;
	RectF		getRenderExtent() const;
	const Rect4F &getRenderExtent4F() const;

	SPACETYPE	getType() const { return spacetype; };
	///@}

	/// @name Space Transform Functions @{
	CoordF transformFromWorld(const CoordF &worldloc);
	CoordF transformToWorld(const Coord &gridcoordinate);
	CoordF transformToWorld(const CoordF &gridcoordinate);
	/// Clip a grid coordinate against the resolution bounds specified
	CoordF clipGridCoord(const CoordF &coord);
	bool   testGridCoord(const Coord &coord);

	/// Convert a grid height to a global height value
	float transformElevation(float elev) const { return vertical.x + vertical.y*elev; };
	/// Convert a world elevation in WM units to grid height
	float transformElevationFromWorld(float elev) const { return (elev-vertical.x) / vertical.y; };

	///@}

	/// ***
	float getDScale() const { if (worldregion.width() > 0.0f) return resolution.x / worldregion.width(); return 0.0f; };

	float getVscale() const { return vertical.y; };
	/// The vertical scaling info is stored as a base,scale pair rather than an elevation pair.
	/// thus to get the upper bound, you will have to do vertical.x + vertical.y
	CoordF getVertical() const { return vertical; };
	/// resolution area
	size_t resarea() const { return size_t(resolution.x) * size_t(resolution.y); };
private:

	inline void createFastMul();

	Coord		resolution;		///< packet area primitives
	Rect4F		worldregion;	///< Region this packet encloses
	SPACETYPE	spacetype;		///< obeys worldspace?
	CoordF		vertical;		///< vertical base,scale pair
private:	
	CoordF		oor;			// one-over-resolution reciprocal for fast transforms
	CoordF		oos;			// one-over-worldsize reciprocal
};

/** BuildContext provides size and build state necessary for a device to build.
	@ingroup data, devices, building

The context information passed includes the preview state, render extents, memory conservation,
 as well as a reporting mechanism to pass progress reports back to the Builder.

BuildContext gets passed to a device's Activate() functions.

BuildContext can be freely shallow-copied to allow for different interior arrangements
	
*/
class BuildContext : public PacketContext {
public:
	BuildContext(BuildResultSet *_result, DeviceWorldBuilder *owner);
	BuildContext(const BuildContext &other);
	BuildContext &operator= (const BuildContext &rhs);

	void assignBuilder(DeviceWorldBuilder *_dwb) { builder = _dwb; };
	void assignResultSet(BuildResultSet *_result) { result_ptr = _result; };

	void			SetWorldSize(RenderExtent &re, WorldCommonData &wpd);

	// Build Progress Reporting operations
	virtual bool ReportDeviceProgress(Device *dev, int amnt);			 ///< Progress on a 0..100 scale
	virtual bool ReportDeviceProgress(Device *dev, int amnt, int max); ///< Progress normalize to 0..max

	// Store or retrieve data from the build context.
	BuildResultSet &results() { return *result_ptr; };
	DeviceResults &device(const Device *dev) const;

	bool setInput(Device *dev, int slot, PacketPointer data);
	bool setOutput(Device *dev, int slot, PacketPointer data);
	bool setParam(Device *dev, int slot, PacketPointer data);
	bool setMask(Device *dev, PacketPointer data);
	bool setCoordPacket(Device *dev, PacketPointer data);
	bool setDefaultOutput(Device *dev, PacketPointer data);

	PacketPointer getCoordPacket(Device *dev);
	ScopedPacketAccess viewDefaultOutput(Device *dev);

	void setBuilt(Device *dev);
	bool isBuilt(Device *dev);	

	bool keepChildResults();		///< current build should keep all internal results for inspection

	bool tiled;						///< flagged if this is a tiled build
	bool conserve_memory;			///< use memory conservation
	bool terminal;					///< Set by builder on a per-device basis. true if this device is a terminal in the current build
	bool allow_device_threading;	///< if multithreading-capable devices should use worker pool threads
	bool export_outputs;			///< outputs should export their contents on this build
	bool pageable_results;			///< if true, build process will mark results as pageable
private:
// Links to the two most important parts of a build context: the builder that is responsible for this context, and the result set of what the build produces
	DeviceWorldBuilder *builder;								///< The Builder that is orchestrating the build sequence, that we're working for
	BuildResultSet *result_ptr;
// Protect 
};
