/** @file Device.h
	Device class definition and many useful constants and helper functions
*/

#pragma once
#ifndef __DEVICE_H
#define __DEVICE_H

#include <vector>
#include "ForwardDef.h"
#include "Interfaces.h"
#include "common.h"
#include "WMPacket.h"
#include "Globals.h"
#include "Parameter.h"
#include "Ports.h"

const int SOCKET_SIZE_X = 8;
const int SOCKET_SIZE_Y = 8;
const int PSOCKET_SIZE_X = 4;
const int PSOCKET_SIZE_Y = 6;

const int DEV_NAME_LEN = 24;

const int DEV_INPUT = 0;
const int DEV_OUTPUT = 1;
const int DEV_PARAMIN = 2;
const int DEV_MASK = 3;
const int DEV_COORD = 4;

/** @name Parameter Reference Macros
	These are defined indirectly so that the same parameter names can be used
	both inside of the device itself, and inside it's custom dialog box, if any.
	@relates Device
*/
///@{
#define ParmIRef(x)		(params.getOwner()->GetIRef(x))
#define ParmFRef(x)		(params.getOwner()->GetFRef(x))
#define ParmBRef(x)		(params.getOwner()->GetBRef(x))
///@}

typedef UINT DEVID;

/**
	Devices are the basic building block that represent action on your data within World Machine.

	A Device takes zero or more inputs, performs some operation,
	and produces zero or more data packets as ouptut. The type of data, and operation, is completely open. 

	Extend the Device class to implement a new terrain action device.
	Typically, you would derive a new Device from one of the subclasses 
	defined in Filter.h or IODevices.h such as Generator, Filter, Combiner, etc.

	@ingroup devices
*/
class Device //: public IDevice
{
public:
	Device(void);
	virtual ~Device(void);

	Device *Clone();		///< Creates an independent copy of this device
	void Destroy();			///< Delete the device using its internally defined deletion operator

	friend class GraphSort;
	friend class DeviceWorldBuilder;
	friend class DeviceWorld;
	friend class GenericDlg;
	friend class BuildEngine;
	friend class PresetManager;

	/// Assign Device to DeviceWorld
	void SetWorld(DeviceWorld *world) { worldptr = world; }; 
	/// Retrieve DeviceWorld
	DeviceWorld *GetWorld() { return worldptr; };	

	/// Query for pointers to advanced interfaces for this device
	virtual WorldMachine::AdvancedDeviceInterface* QueryInterface(WorldMachine::ADVANCED_INTERFACE_TYPE queryInt);

///@name Serialization Functions
/// Uses C++ streams for IO.
//@{
	virtual bool Load(std::istream &in) =0;
	virtual bool Save(std::ostream &out) =0;
//@}

/// @name UI functions
/// All of these are used internally by the World Machine UI.
	//@{
	virtual dword Color() { return 0x07f7f7f; }; 
	virtual int GetWidth();
	virtual int GetHeight();
	virtual char *GetDescription() =0 { return "This part can never be instanced";};
	virtual void  SetName(const char *name);
	virtual char *GetName();
	virtual char *GetTypeName() =0 { return "Unknown Device"; };

	virtual char *GetInputName(int slot);
	virtual char *GetOutputName(int slot);
	char *GetParamInputName(int slot);		// Non-overridable since it directly interfaces with parameterset

	/// Get location in Device workview
	Coord GetLoc() { return loc; };
	/// Set UI location
	void SetLoc(Coord &newloc) { loc = newloc; }; 
	int	FindPort(Coord loc, int *io, int showstate); /// Discover the port present at a given location
	Coord GetPortLoc(int slot, int io);	///< Retrieve coordinates of a given port, relative to device origin
	Coord GetPortLoc(Port *target_port, int io = -1);	///< Retrieve coordinates of a pointer-to-Port
	//@}

	/// Get Device nametag
	char *GetStaticTag() { return lifeptrs.nametag; };	
	virtual int GetDeviceFamily() = 0;	///< Get Device family (used both for UI and crude RTTI)
	virtual void AssignNewDeviceID();	///< Assigns this device a new UID.
	/// Assign Device a new UID. This is DANGEROUS as you could setup a UID collision; Refrain unless you REALLY know what you're doing.
	void SetDeviceID( int new_id) { device_uid = new_id; }; 
	/// Get Device's UID.
	int GetDeviceID() const { return device_uid; }; 

/// Layout View functions (optional!)
	virtual LayoutUIClient *GetLayoutUI() { return NULL; };

/** @name Device Origin / Coordinate System Transformation functions	
	@deprecated Most of these functions have now been deprecated and moved to Context
*/
	//@{
	/// Return device origin, modified by the coordinate input, if any
	CoordF getOrigin();						
	CoordF getOrigin(BuildContext &context);
	/// return unmodified device origin
	CoordF getOriginRaw() { return origin; };
	void   setOrigin(CoordF &o) { origin = o; };
	/// If this device operates in world or local space
	int	   getOriginType() { return origin_type; };	
	/// Set worldspace behavior of the device
	void   setOriginType(int t) { origin_type = t; };
	//@}


/** @name World Context operations
	These functions wrap and provide convenient access the device build results that are actually properly stored in the DeviceWorld. 
*/
	virtual PacketPointer	getDefaultOutput(PSTATE preview);
	virtual Device *		getDisplayDevice();
	virtual PacketPointer	getFinishedPacket(PSTATE state, int slot = 0);
	ScopedPacketAccess		viewInput(int slot, PSTATE preview);

/// @name Graph&Port operations
	//@{
	void LinkClear();										///< Clear all links to/from this device
	Port *GetPort(int index, int io);						///< Return a pointer to the given port
	int  GetNextPort(int io);								///< Returns the next unconnected port index of a type, or -1 if no unconnected ports of that type exist
	bool CheckPort(int slot, int io);						///< Is the specified port connected?
	int  GetPortStyle(int slot, int io);					///< Return if the port is flagged special or not
	bool FindPort(const Port *port, int *index, int *io);	///< Look for the specified port and return its slot number and IO type on THIS device, if it exists
	int  GetLinkedPort(Port *linked_port, int io = -1);		///< Given a port on a DIFFERENT device, return the slot number of the connected port on THIS device

	/// Safely retrieve one of a device's input ports
	Port *in(int slot) { if (slot < (int) inputs.size()) return inputs[slot]; else return NULL;};
	/// Safely retrieve one of a device's output ports
	OutputPort *out(int slot) { if (slot < (int) outputs.size()) return outputs[slot]; else return NULL; };
	/// Safely retrieve one of a device's parameter ports
	Port *parmin(int slot) { if (slot < (int) params_input.size()) return params_input[slot]; else return NULL; };
	/// Return the mask port
	Port *mask() { return mask_input; };	
	/// Return the coordinate port
	Port *coord_port() { return coord_input; };	

	int  highest_input();	/// Retrieve index of the highest connected INPUT. -1 == no connections
	int  highest_output();	/// Retrieve index of the highest connected OUTPUT
	int  active_inputs();	/// number of connected inputs
	int  active_outputs();	/// number of connected outputs
	int  active_param_inputs();	/// number of connected param inputs
	/// Number of inputs
	int	 max_inputs() const { return (int) inputs.size(); };
	/// Number of outputs
	int  max_outputs() const { return (int) outputs.size(); };
	/// Number of parameter slots
	int  max_parm() const	{ return (int) params_input.size(); };
	//@}

/// @name Device functionality
/// These are the heart of the device, and control both the actual effect and the interaction with the Device.
	//@{
	virtual bool RunUI();				///< Run the user interface for this part. Over-ride to show a custom dlg
	bool		 RunNameDialog();		///< Run the user interface to set a new name for this part
	virtual void RunHelpDialog();		///< Run the user interface to display a dialog with info about the device

	virtual ACTIVATE_RESULT ActivateSequence(BuildContext &context);	///< Wraps call to PreAct, ActOnNeed, and PostAct
	virtual ACTIVATE_RESULT ActivateOnNeed(BuildContext &context);		///< Calls Activate only if changed. Generally don't override this
protected:
	virtual bool Activate(BuildContext &context) = 0;			///< THE ALL IMPORTANT Activate call. Actually PERFORM the device action!
public:

	virtual bool RunAuxAction(int index);		///< Run extra actions defined by this sprocket. Returns TRUE if framework should re-calculate after action
	virtual int  getAuxActionSize();			///< The number of actions defined
	virtual char *getAuxActionName(int index);	///< The name of the extra action

	//@}

	DEV_DATA &GetLifeVars() { return lifeptrs; };
	virtual bool Randomize(int newseed = 0);		///< Randomize this device, if applicable.

	void Preview();
	virtual void Falsify();	///< Inform parent world that we have been falsified

/// @name Parameter Controls. The reference versions are allowed to throw exceptions
/// @throws WMException
	//@{
	int		GetI( const char *name);
	float	GetF( const char *name);
	bool	GetB(const char *name);

	int		&GetIRef( const char *name);		///< Get a reference to the given integer parameter
	float	&GetFRef( const char *name);		///< Get a reference to the given float parameter
	bool	&GetBRef(const char *name);			///< Get a reference to the given boolean parameter
	dword   &GetCRef(const char *name);			///< Get a reference to the given RGB Color parameter

	ParameterSet &GetParams() { return params; };
	void SetParams(ParameterSet &other) { params = other; }; //params.setOwner(this); };
	PacketPointer RetrieveParamPacket( int slot );
	//@}


	/// Notification/Event enums
	// World events that a device may or may not depend on. Right now just a few of these...
	// used by isUpdatedOn to control if the device is reset or not on this event
	enum DEVICE_EVENT { DEV_EVENT_LIGHTING = 1, DEV_EVENT_WATER = 2, DEV_EVENT_COLORTABLE = 4, DEV_EVENT_RENDER_EXTENTS = 8 };

	/// @name Notification Functions
	/// Provide a means for a Device to be told about certain events happening in the world. This can be useful both internally (things like
	/// morphing output ports, etc) and externally (synchronizing with another application)
	///
	/// Notifications such as PortConnected, Added, etc are sent whenever they occur, even if that means they occur during world loading.
	//@{

	/// Notification that this device has been added to a world. This will occur after a file is fully loaded, or immediately after a device
	/// is added to an already-existing world. The loaded parameter indicates if this was from a file load or simply a part added.
	virtual void notifyAdded(bool loaded);

	/// Notification that this device has been removed from a world and is about to be destroyed
	virtual void notifyRemoved();

	/// Notification that this device's world is about to be closed. This occurs when a file is closed or when World Machine exits
	virtual void notifyWorldClosing();

	/// Notification when a wire is attached or removed to a port for this device.
	/// You will receive the notification AFTER the event in question has occured; this may be useful for doing things like morphing port type, etc.
	virtual void notifyPortConnected(int slot, int io, bool is_connecting);

	/// Notification that this device has been falsified, and its contents have been removed/are no longer valid.
	virtual void notifyFalsify();

	// Notification that some world settings have been changed. The specific type of change is specified by the event parameter
	virtual void notifyWorldChanged(DEVICE_EVENT event, VAL value);

	/// Notification sent just before a final-mode build begins. the tiled boolean indicates if it is a tiled or regular build
	/// Note that you will receive this before any devices begin building
	virtual void notifyBuildStarting(bool tiled);

	/// Notification sent when a build has completed. the tiled boolean indicates if it is a tiled or regular build
	/// You will receive this when all devices that are queued to build are complete.
	/// Note that for a Tiled Build, this means that all tiled have been blended as well.
	virtual void notifyBuildEnd(bool tiled);

	//@}

	/// @name Context Sensitive Queries
	/// These are convenience wrappers around the objects world state tat is actually stored in the passed-in buildContext
	bool isBuilt(PSTATE state);
	bool isBuilt(BuildContext &context);
	/// Is this device asleep (signal'd to not build because of child caching
	bool		 isAsleep() { return false; }; //{ return (processed & DEV_SLEEPING) != 0; };	

	/// @name Interrogatives
	/// Dynamic state of the device
	//@{
	/// device is ghosted if it doesn't have a current owner world
	bool		 isGhosted() { return !worldptr; };	
	/// Is this device toggled to disable mode?
	bool		 isEnabled() { return enable; };	
	/// Is this device bypassed?
	bool		 isBypassed() { return bypass; };	
	//@}
	/// enable/disable this device. A disabled device does not produce any output
	void		setEnable(bool _en) { enable = _en; };	
	/// set device to be bypassed or not. A bypassed device simply passes its inputs to its outputs with no further processing
	void		setBypass(bool _by) { bypass = _by; };	

	/// @name Intrinsic Interrogatives
	/// These reflect the capabilities of the device class. Your class should implement these if your answers are different from the defaults.
	//@{
	/// Device returns true if it needs to be recreated when the given world parameter changes
	virtual bool isUpdatedOn(int) { return false; }; 
	/// If this device requires the registered version or not.
	virtual bool isRegistered() { return false; };	
	/// This device has a sensible output for viewing? 
	virtual bool isViewable() { return true; };		
	/// Should this device be unenableable & undeletable in the workview?
	virtual bool isPermanent() { return false; };	
	/// Allow presets for this device in the generic dialog manager?
	virtual bool isPresetCapable() { return true; };
	/// If this device is compatible with explorer mode or not
	virtual bool isExplorable() { return true; };   
	/// Should this device have a mask input?
	virtual bool isMaskable() { return false; };		
	/// Does this device use the origin value?
	virtual bool hasOrigin() { return false; };		
	/// Does this device allow local space?
	virtual bool hasLocalOverride() { return false; }; 
	/// Is this device a building checkpoint
	virtual bool isCheckPoint() { return false; };	
	//@}

	/// @name Hinting
	//@{
	virtual int getSpeedClass() { return 3; };		// Provide a rough estimate of processing speed of this device, relative to perlin noise (3). Higher is slower
	virtual int  estimateMemoryUsage();				// Provide the framework with an estimate of the memory required by this device given current settings, in KB

	void setMaskOverride(int value = 0);	// Temporarily override how this device outputs data
	int  getMaskOverride() { return mask_override; };
	//@}

protected:
	virtual void doClone(Device *other);									/// Copy data from other to this

	// override to control how we mask against inputs. Default does no masking.
	virtual bool doMaskCreateNewCopies(std::vector<PacketPointer> &results, BuildContext &context);
	void doMaskOperation(PacketPointer _result, PacketPointer _original, PacketPointer _mask);	// Actual blending operation

/// @name Data input/output/activation functions
/// Note that in our current model these are simply convenience wrappers to the appropriate BuildContext functions
	//@{
	void StoreData(PacketPointer const&dat, int slot, BuildContext &context);			// Stores a completed HF in our output buffer
	PacketPointer RetrieveInput(int slot, BuildContext &context);				// Retrieve main inputs data, and TRANSFERS ownership
	//@}

	bool Bypass(BuildContext &context);						///< Perform a bypass of this device, if possible

	DEV_DATA	lifeptrs;			///< make&destroy pointers for this device
	DeviceWorld *worldptr;			///< pointer to this device's container
	Coord loc;						///< Coordinate of this device in world space

	bool enable;					///< If this device is disabled from activation or not.
	bool bypass;					///< If this device should be bypassed
	int  mask_override;				///< 0 = normal masking
									///< 1 = override mask on
									///< 2 = override mask off

	void SetLinks(int in, int out, int packet_type = WMP_hfield);	 // Setup this device's link vector
	void SetLinksFlags(int in, int out);// Setup port flags
	void ChangeLinks(int in, int out, int packet_type = WMP_hfield);// Change the # of links without destroying existing ones. Used by devices
	void ChangeLinkIn(int in, int packet_type = WMP_hfield);		// for on-the-fly modifications of the # of links
	void ChangeLinkOut(int out, int packet_type = WMP_hfield);
	void MorphOutput(int slot, int packet_type);					///< Change the type of a single link, breaking any existing links that are no longer of the right type

	virtual void AddPort(int flag, int io, int packet_type = WMP_hfield, int index = -1);
	virtual void RemovePort(int io, int index );


	void AddParam(Parameter &p, int flags = 0);	///< Adds a parameter to paramset, and also creates an override port for it
	void RemoveParam(int index);				///< Remove a parameter, delete its override port
	void ResetParams();							///< Deletes all params and param ports
	ParameterSet			params;				///< Set of user controllable, managed-parameters
 
	std::vector<Port*>		inputs;			///< Input ports
	std::vector<OutputPort*>outputs;		///< Output ports 
	std::vector<Port*>		params_input;	///< Parameter-override hookups

	Port *mask_input;				///< Built-in mask input
	Port *coord_input;				///< Built-in coordinate input

	char name[DEV_NAME_LEN];		///< name of this device
private:
	CoordF origin;					///< the device's origin.
	int	origin_type;				
	DEVID device_uid;			///< Device unique identifier. NOT globally unique - guarenteed unique only within the current world file.


	void ChangedLinks();

	int doStoreDistribute(BuildContext &context);
	PacketPointer reachThroughInput(BuildContext &context, Port *port);
};

#endif
