#pragma once
/// @file WMException.h


/// @name Exception Codes
/// @relates WMException
//@{
const int EXCEPTION_OUT_OF_MEMORY = 1;
const int EXCEPTION_BADDEVICETAG = 2;
const int EXCEPTION_TILED_BUILD_FILE_ERROR = 3;
const int EXCEPTION_LAYOUT_EXCEPTION = 4;
const int EXCEPTION_TEXT_EXCEPTION = 5;
const int EXCEPTION_LOAD_ERROR = 6;
const int EXCEPTION_GENERIC = 7;
const int EXCEPTION_BAD_PARAMETER = 8;
const int EXCEPTION_TILED_BUILD_ERROR = 9;
//@}

namespace WorldMachine {
	/**
	 	WMException is the base of the exception heirarchy in World Machine. Please see \Exceptions and World Machine for more details about Exception handling in WM.

		In general we make very limited use of Exceptions, principally limited to signaling out of memory conditions and malformed inputs/outputs that should not occur during production.
		Due to the plugin nature of World Machine, exceptions are often prevented from propagating fully up the call stack to make sure they do not cross DLL boundaries. See the above for more info.
	 */
	class WMException : public std::runtime_error {
	public:
		WMException()  : std::runtime_error(nullptr) {};
		WMException(const char* _str) : std::runtime_error(_str) {};
	};

	class OutOfMemoryException : public WMException {
	public:
		OutOfMemoryException(size_t requested_memory_size, size_t current_allocated_size) {
			request = requested_memory_size; memstate = current_allocated_size;
		}

		virtual const char *what() const { return "memory allocation failure" ;};
	private:
		size_t request, memstate;
	};

	/**
	 	Throw when a DeviceBuilder, ThreadedDeviceBuilder, or similar encounters an exception on a different thread of execution. The exception here represents ane xception thrown
		in a different thread.
	 */
	class BuildException : public WMException {
		virtual const char *what() const { return "repackaged exception failure during build, or miscellaneous build failure." ;};
	};

	/**
	 	Throw when a device or component has an IO failure. You can optionally specify a message
	 */
	class ComponentIOException : public WMException {
	public:
		ComponentIOException(const char* msg) : WMException(msg) {};
		virtual const char *what() const { return "A device or deviceworld could not load a necessary component. Possible corrupted file." ;};
	};

	/** Throw from devices or components when a version being loaded is incompatible
	 **
	 */
	class ComponentInputVersionException : public ComponentIOException
	{
	public:
		ComponentInputVersionException(const char*msg, int in_ver) : ComponentIOException(msg) { ver = in_ver; };
		virtual const char *what() const { return "A device or deviceworld encountered a version number that it does not know how to handle. This file might be from a new or very old version of World Machine" ;};
	private:
		int ver; 
	};

	/**
	 	Thrown when a device parameter is requested that does not exist in the ParameterSet.
	 */
	class ParameterException : public WMException {
	public:
		ParameterException(const char *whatparam) : WMException(whatparam) {};
		virtual const char *what() const { return "Parameter specified does not exist!" ;};
	};

	/**
	 	Thrown when a file cannot be written to or accessed.
	 */
	class FileAccessException : public WMException {
	public:
		FileAccessException(const char*filename) : WMException(filename) {};
		virtual const char *what() const { return "The filename specified could not be accessed."; };
	private:
	};

	class StopException : public WMException {
		virtual const char *what() const { return "Stop requested while processing."; };
	};
}
